@@ -1,13 +1,25 @@
-import warnings
-import zipfile
 import io
 import json
-from .base import VulnersApiBase, ParamError, validate_params
-from .base import Endpoint, String, Integer, Dict, List, Tuple, Const, ResultSet, Boolean
+import warnings
+import zipfile
+
+from .base import (
+    Boolean,
+    Const,
+    Dict,
+    Endpoint,
+    Integer,
+    List,
+    ParamError,
+    ResultSet,
+    String,
+    Tuple,
+    VulnersApiBase,
+    validate_params,
+)
 
 
 class VulnersApi(VulnersApiBase):
-    search_size = 100
     default_fields = (
         "id",
         "title",
@@ -22,6 +34,8 @@ class VulnersApi(VulnersApiBase):
         "sourceHref",
         "sourceData",
         "cvelist",
+        "vulnStatus",
+        "assigned",
     )
 
     __search = Endpoint(
@@ -38,7 +52,7 @@ class VulnersApi(VulnersApiBase):
     @validate_params(
         query=String(),
         limit=Integer(minimum=1, maximum=100),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         fields=Tuple(item=String()),
     )
     def find(self, query, limit=20, offset=0, fields=default_fields):
@@ -53,21 +67,14 @@ class VulnersApi(VulnersApiBase):
         Returns list of the documents.
         Use .total to get the total number of found documents.
         """
-        end = min(10000, offset + limit)
-        if offset >= end:
-            return ResultSet.from_dataset(
-                [], self.__search(query, 0, 1, ["id"])["total"]
-            )
-        else:
-            search = self.__search(query, offset, limit, fields)
-            return ResultSet.from_dataset(
-                [e["_source"] for e in search["search"]], search["total"]
-            )
+        limit = min(limit, 10000 - offset)
+        search = self.__search(query, offset, limit, fields)
+        return ResultSet.from_dataset([e["_source"] for e in search["search"]], search["total"])
 
     @validate_params(
         query=String(),
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         fields=Tuple(item=String()),
     )
     def find_all(self, query, limit=20, offset=0, fields=default_fields):
@@ -82,31 +89,30 @@ class VulnersApi(VulnersApiBase):
         Returns list of the documents.
         Use .total to get the total number of found documents.
         """
-        end = min(10000, offset + limit)
-        if offset >= end:
-            return ResultSet.from_dataset(
-                [], self.__search(query, 0, 1, ["id"])["total"]
-            )
+        limit = min(limit, 10000 - offset)
+        end = offset + limit
+        batch_size = min(1000, limit)
         result = ResultSet()
-        batch_size = min(self.search_size, limit)
-        for skip in range(offset, end, batch_size):
-            chunk = self.__search(query, skip, min(batch_size, end - skip), fields)
-            result += [e["_source"] for e in chunk["search"]]
+        while len(result) < limit:
+            chunk = self.__search(query, offset, min(batch_size, end - offset), fields)
+            data = chunk["search"]
+            result += [e["_source"] for e in data]
             result.total = chunk["total"]
-            if len(chunk["search"]) < batch_size:
+            if not data:
                 break
+            if result.total <= len(data):
+                break
+            offset += len(data)
         return result
 
     @validate_params(
         query=String(),
         lookup_fields=Tuple(item=String()),
         limit=Integer(minimum=1, maximum=100),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         fields=Tuple(item=String()),
     )
-    def find_exploit(
-        self, query, lookup_fields=None, limit=20, offset=0, fields=default_fields
-    ):
+    def find_exploit(self, query, lookup_fields=None, limit=20, offset=0, fields=default_fields):
         """
         Search in Vulners database for the exploits.
 
@@ -130,8 +136,8 @@ class VulnersApi(VulnersApiBase):
     @validate_params(
         query=String(),
         lookup_fields=Tuple(item=String()),
-        limit=Integer(minimum=1, maximum=100),
-        offset=Integer(minimum=0, maximum=10000),
+        limit=Integer(minimum=1, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         fields=Tuple(item=String()),
     )
     def find_exploit_all(
@@ -170,41 +176,87 @@ class VulnersApi(VulnersApiBase):
     __get_burp_software = Endpoint(
         method="post",
         url="/api/v3/burp/softwareapi/",
-        params=[("software", String()), ("version", String()), ("type", String()), ("exactmatch", Boolean(default=False))],
+        params=[
+            ("software", String()),
+            ("version", String(required=False)),
+            ("vendor", String(required=False)),
+            ("update", String(required=False)),
+            ("language", String(required=False)),
+            ("sw_edition", String(required=False)),
+            ("target_sw", String(required=False)),
+            ("target_hw", String(required=False)),
+            ("respect_major_version", String(required=False, choices=["yes", "no", "true", "false"])),
+            ("exclude_any_version", String(required=False, choices=["yes", "no", "true", "false"])),
+            ("type", String(required=False)), # deprecated
+            ("exactmatch", Boolean(default=False)), # deprecated
+        ],
         content_handler=_get_burp_software_content,
     )
 
     del _get_burp_software_content
 
     @validate_params(name=String(), version=String())
-    def get_software_vulnerabilities(self, name, version):
+    def get_software_vulnerabilities(
+        self,
+        name,
+        version,
+        vendor=None,
+        update=None,
+        language=None,
+        sw_edition=None,
+        target_sw=None,
+        target_hw=None,
+        respect_major_version=None,
+        exclude_any_version=None,
+        only_ids=None
+    ):
         """
         Find software vulnerabilities using name and version.
 
         name: Software name, e.g. 'httpd'
         version: Software version, e.g. '2.1'
         """
-        return self.__get_burp_software(name, version, "software")
+        return self.__get_burp_software(
+            name,
+            version,
+            vendor,
+            update,
+            language,
+            sw_edition,
+            target_sw,
+            target_hw,
+            respect_major_version,
+            exclude_any_version,
+            only_ids
+        )
 
     @validate_params(cpe=String())
-    def get_cpe_vulnerabilities(self, cpe, exactmatch=False):
+    def get_cpe_vulnerabilities(
+        self,
+        cpe,
+        respect_major_version=None,
+        exclude_any_version=None,
+        only_ids=None
+    ):
         """
         Find software vulnerabilities using CPE string. See CPE references at https://cpe.mitre.org/specification/
 
         cpe: CPE software string, see https://cpe.mitre.org/specification/
         exactmatch:  if true searches only for bulletins corresponding to the specified minor version and revision
         """
-        cpe_split = cpe.split(":")
-        if len(cpe_split) <= 4:
-            raise ParamError("Malformed %s", "cpe")
-        if cpe_split[1] == '2.3':
-            version = cpe_split[5]
-        elif cpe_split[1] in '/a/o/h':
-            version = cpe_split[4]
-        else:
-            raise ParamError("Malformed %s", "cpe")
-
-        return self.__get_burp_software(cpe, version, "cpe", exactmatch=exactmatch)
+        return self.__get_burp_software(
+            cpe,
+            None,
+            None,
+            None,
+            None,
+            None,
+            None,
+            None,
+            respect_major_version,
+            exclude_any_version,
+            only_ids,
+        )
 
     get_multiple_bulletins = Endpoint(
         method="post",
@@ -213,9 +265,7 @@ class VulnersApi(VulnersApiBase):
         params=[
             (
                 "id",
-                Tuple(
-                    item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"
-                ),
+                Tuple(item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"),
             ),
             ("fields", Tuple(item=String(), default=default_fields)),
         ],
@@ -238,9 +288,7 @@ class VulnersApi(VulnersApiBase):
         params=[
             (
                 "id",
-                List(
-                    item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"
-                ),
+                List(item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"),
             ),
             ("fields", Tuple(item=String(), default=default_fields)),
             ("references", Const(True)),
@@ -257,6 +305,29 @@ class VulnersApi(VulnersApiBase):
         """
         return self.get_multiple_bulletin_references([id], fields=fields).get(id, {})
 
+    get_multiple_documents_with_references = Endpoint(
+        method="post",
+        url="/api/v3/search/id/",
+        description="",
+        params=[
+            (
+                "id",
+                List(item=String(), description="List of ID's. E.g., ['CVE-2017-14174']"),
+            ),
+            ("fields", Tuple(item=String(), default=default_fields)),
+            ("references", Const(True)),
+        ],
+    )
+
+    @validate_params(id=String(), fields=Tuple(item=String()))
+    def get_document_with_references(self, id, fields=default_fields):
+        """
+        Fetch bulletin with references by identificator
+
+        identificator: Bulletin ID. E.g., "CVE-2017-14174"
+        """
+        return self.get_multiple_documents_with_references([id], fields=fields)
+
     @validate_params(kbid=String())
     def get_kb_seeds(self, kbid):
         """
@@ -296,9 +367,7 @@ class VulnersApi(VulnersApiBase):
         params=[
             (
                 "os",
-                String(
-                    description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"
-                ),
+                String(description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"),
             ),
             ("version", String(description="OS version")),
             (
@@ -323,9 +392,7 @@ class VulnersApi(VulnersApiBase):
         params=[
             (
                 "os",
-                String(
-                    description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"
-                ),
+                String(description="Full name of the OS. Like Ubuntu, Debian, rhel, oraclelinux"),
             ),
             ("version", String(description="OS version", param="osVersion")),
             ("packages", List(item=Dict(), description="List of the software dicts")),
@@ -365,11 +432,14 @@ class VulnersApi(VulnersApiBase):
                 "os_version",
                 String(description="Windows OS version, like '10.0.19045'"),
             ),
+            (
+                "platform",
+                String(required=False, description="os platform like x86"),
+            ),
             (
                 "kb_list",
                 List(
-                    item=String(),
-                    description="List of installed KB's, ['KB2918614', 'KB2918616']"
+                    item=String(), description="List of installed KB's, ['KB2918614', 'KB2918616']"
                 ),
             ),
             (
@@ -440,9 +510,7 @@ class VulnersApi(VulnersApiBase):
     del _unpack_json_file
 
     @validate_params(collection=String(), start_date=String(), end_date=String())
-    def get_collection(
-        self, collection, start_date="1950-01-01", end_date="2199-01-01"
-    ):
+    def get_collection(self, collection, start_date="1950-01-01", end_date="2199-01-01"):
         """
         Get entire collection data
 
@@ -450,12 +518,8 @@ class VulnersApi(VulnersApiBase):
         """
         collections = self.get_suggestion("type")
         if collection not in collections:
-            raise ParamError(
-                "Unknown %%s. Available values are %s" % (collections,), "collection"
-            )
-        return self.__archive_collection(
-            type=collection, datefrom=start_date, dateto=end_date
-        )
+            raise ParamError("Unknown %%s. Available values are %s" % (collections,), "collection")
+        return self.__archive_collection(type=collection, datefrom=start_date, dateto=end_date)
 
     @validate_params(os=String(), version=String())
     def get_distributive(self, os, version):
@@ -467,9 +531,7 @@ class VulnersApi(VulnersApiBase):
         """
         supported_os = self.get_suggestion("affectedPackage.OS")
         if os.lower() not in [os_name.lower() for os_name in supported_os]:
-            raise ParamError(
-                "Unknown %%s. Available values are %s" % (supported_os,), "os"
-            )
+            raise ParamError("Unknown %%s. Available values are %s" % (supported_os,), "os")
         data = self.__distributive(os=os, version=version)
         return [bulletin["_source"] for bulletin in data]
 
@@ -477,22 +539,38 @@ class VulnersApi(VulnersApiBase):
         method="post",
         url="/api/v3/reports/vulnsreport",
         params=[
-            ("reporttype", String(description="One of strings [vulnssummary, vulnslist, ipsummary, scanlist]")),
+            (
+                "reporttype",
+                String(description="One of strings [vulnssummary, vulnslist, ipsummary, scanlist]"),
+            ),
             ("skip", Integer(description="Skip this amount of items. 10000 is the hard limit")),
-            ("size", Integer(description="The maximum number of items to return. 10000 is the hard limit")),
-            ("filter", Dict(description="Dict of fields to filter, eg { 'OS': 'Centos', 'OSVersion': '7'}")),
-            ("sort", String(description="Field to sort, eg 'severity' or '-severity' to sort desc")),
+            (
+                "size",
+                Integer(
+                    description="The maximum number of items to return. 10000 is the hard limit"
+                ),
+            ),
+            (
+                "filter",
+                Dict(
+                    description="Dict of fields to filter, eg { 'OS': 'Centos', 'OSVersion': '7'}"
+                ),
+            ),
+            (
+                "sort",
+                String(description="Field to sort, eg 'severity' or '-severity' to sort desc"),
+            ),
         ],
-        content_handler=lambda x, _: x['report']
+        content_handler=lambda x, _: x["report"],
     )
 
     @validate_params(
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         filter=Dict(),
-        sort=String()
+        sort=String(),
     )
-    def vulnssummary_report(self, limit=30, offset=0, filter=None, sort=''):
+    def vulnssummary_report(self, limit=30, offset=0, filter=None, sort=""):
         """
         Get Linux Audit results. Return summary for all found vulnerabilities - id, title, score, severity etc
 
@@ -506,11 +584,11 @@ class VulnersApi(VulnersApiBase):
 
     @validate_params(
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         filter=Dict(),
-        sort=String()
+        sort=String(),
     )
-    def vulnslist_report(self, limit=30, offset=0, filter=None, sort=''):
+    def vulnslist_report(self, limit=30, offset=0, filter=None, sort=""):
         """
         Get Linux Audit results. Return list of vulnerabilities found on hosts:
             vulnerability id, vulnerability title, vulnerability severity, host information,  etc
@@ -525,11 +603,11 @@ class VulnersApi(VulnersApiBase):
 
     @validate_params(
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         filter=Dict(),
-        sort=String()
+        sort=String(),
     )
-    def ipsummary_report(self, limit=30, offset=0, filter=None, sort=''):
+    def ipsummary_report(self, limit=30, offset=0, filter=None, sort=""):
         """
         Get Linux Audit results. Return summary for hosts:
             agent id, host ip, host fqdn, os name and version, found vulnerabilities count and severity
@@ -544,11 +622,11 @@ class VulnersApi(VulnersApiBase):
 
     @validate_params(
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         filter=Dict(),
-        sort=String()
+        sort=String(),
     )
-    def scanlist_report(self, limit=30, offset=0, filter=None, sort=''):
+    def scanlist_report(self, limit=30, offset=0, filter=None, sort=""):
         """
         Get Linux Audit results. Return list of scans:
            host ip and fqdn, os name and version, scan date, cvss score
@@ -561,10 +639,47 @@ class VulnersApi(VulnersApiBase):
 
         return self.__report("scanlist", offset, limit, filter or {}, sort)
 
-    get_webhooks = Endpoint(
+    get_subscriptions = Endpoint(
+        method="get",
+        url="/api/v3/subscriptions/listEmailSubscriptions/",
+        content_handler=lambda hooks, _: hooks["subscriptions"],
+    )
+
+    add_subscription = Endpoint(
         method="post",
+        url="/api/v3/subscriptions/addEmailSubscription/",
+        params=[
+            ("query", String()),
+            ("email", String()),
+            ("format", String(default="html", choices=("html", "json", "pdf"))),
+            ("crontab", String(allow_null=True, default=None)),
+            ("query_type", String(default="lucene")),
+        ]
+    )
+
+    edit_subscription = Endpoint(
+        method="post",
+        url="/api/v3/subscriptions/editEmailSubscription/",
+        params=[
+            ("subscriptionid", String()),
+            ("format", String(allow_null=True, default=None, choices=("html", "json", "pdf"))),
+            ("crontab", String(allow_null=True, default=None)),
+            ("active", String(allow_null=True, default=None, choices=("yes", "no", "true", "false"))),
+        ]
+    )
+
+    delete_subscription = Endpoint(
+        method="post",
+        url="/api/v3/subscriptions/removeEmailSubscription/",
+        params=[
+            ("subscriptionid", String()),
+        ]
+    )
+
+    get_webhooks = Endpoint(
+        method="get",
         url="/api/v3/subscriptions/listWebhookSubscriptions/",
-        content_handler=lambda hooks, _: hooks["subscriptions"]
+        content_handler=lambda hooks, _: hooks["subscriptions"],
     )
 
     add_webhook = Endpoint(
@@ -574,7 +689,7 @@ class VulnersApi(VulnersApiBase):
             ("query", String()),
             ("active", Boolean(default=True)),
         ],
-        content_handler=lambda hook, _: hook["subscription"]
+        content_handler=lambda hook, _: hook["subscription"],
     )
 
     enable_webhook = Endpoint(
@@ -583,7 +698,7 @@ class VulnersApi(VulnersApiBase):
         params=[
             ("subscriptionid", String()),
             ("active", Boolean()),
-        ]
+        ],
     )
 
     delete_webhook = Endpoint(
@@ -591,25 +706,22 @@ class VulnersApi(VulnersApiBase):
         url="/api/v3/subscriptions/deleteWebhookSubscription/",
         params=[
             ("subscriptionid", String()),
-        ]
+        ],
     )
 
     read_webhook = Endpoint(
         method="get",
         url="/api/v3/subscriptions/webhook",
-        params=[
-            ("subscriptionid", String()),
-            ("newest_only", String(default="true"))
-        ]
+        params=[("subscriptionid", String()), ("newest_only", String(default="true"))],
     )
 
     @validate_params(
         limit=Integer(minimum=1, maximum=10000),
-        offset=Integer(minimum=0, maximum=10000),
+        offset=Integer(minimum=0, maximum=9999),
         filter=Dict(),
-        sort=String()
+        sort=String(),
     )
-    def hostvulns_report(self, limit=30, offset=0, filter=None, sort=''):
+    def hostvulns_report(self, limit=30, offset=0, filter=None, sort=""):
         """
         Get Linux Audit results. Return list of hosts and host vulnerabilities:
            host ip and fqdn, os name and version, cumulative fix, vulnerability ids
@@ -622,6 +734,18 @@ class VulnersApi(VulnersApiBase):
 
         return self.__report("hostvulns", offset, limit, filter or {}, sort)
 
+    @validate_params(id=String())
+    def get_bulletin_history(self, id):
+        data = self.__get_bulletin_history(id=id)
+        return data["result"]
+
+    __get_bulletin_history = Endpoint(
+        method="get",
+        url="/api/v3/search/history/",
+        description="Bulletin history list",
+        params=[("id", String(description="Bulletin ID"))],
+    )
+
 
 _Unset = object()
 
@@ -652,13 +776,9 @@ class DeprecatedVulnersApi(VulnersApi):
             "Use VulnersApi.find() instead.",
             DeprecationWarning,
         )
-        return self.find(
-            query, limit=pageSize, offset=offset, fields=fields or self.default_fields
-        )
+        return self.find(query, limit=pageSize, offset=offset, fields=fields or self.default_fields)
 
-    def searchExploit(
-        self, query, lookup_fields=None, limit=100, offset=0, fields=None
-    ):
+    def searchExploit(self, query, lookup_fields=None, limit=100, offset=0, fields=None):
         warnings.warn(
             "searchExploit() is deprecated and will be removed in future release. "
             "Use VulnersApi.find_exploit_all() instead.",
@@ -675,9 +795,7 @@ class DeprecatedVulnersApi(VulnersApi):
             fields=fields or self.default_fields,
         )
 
-    def searchExploitPage(
-        self, query, lookup_fields=None, limit=100, offset=0, fields=None
-    ):
+    def searchExploitPage(self, query, lookup_fields=None, limit=100, offset=0, fields=None):
         warnings.warn(
             "searchExploitPage() is deprecated and will be removed in future release. "
             "Use VulnersApi.find_exploit() instead.",
@@ -750,9 +868,7 @@ class DeprecatedVulnersApi(VulnersApi):
             "Use VulnersApi.get_multiple_bulletins() instead.",
             DeprecationWarning,
         )
-        return self.get_multiple_bulletins(
-            identificatorList, fields=fields or self.default_fields
-        )
+        return self.get_multiple_bulletins(identificatorList, fields=fields or self.default_fields)
 
     def references(self, identificator, fields=None):
         warnings.warn(
@@ -760,9 +876,7 @@ class DeprecatedVulnersApi(VulnersApi):
             "Use VulnersApi.get_bulletin_references() instead.",
             DeprecationWarning,
         )
-        return self.get_bulletin_references(
-            identificator, fields=fields or self.default_fields
-        )
+        return self.get_bulletin_references(identificator, fields=fields or self.default_fields)
 
     def referencesList(self, identificatorList, fields=None):
         warnings.warn(
@@ -820,9 +934,7 @@ class DeprecatedVulnersApi(VulnersApi):
             "Use VulnersApi.get_kb_updates() instead.",
             DeprecationWarning,
         )
-        return self.get_kb_updates(
-            kb_identificator, fields=fields or self.default_fields
-        )
+        return self.get_kb_updates(kb_identificator, fields=fields or self.default_fields)
 
     def autocomplete(self, query):
         warnings.warn(
